/**
  *  \file client/tiles/planetcolonist.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/planetcolonist.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "util/misc.h"
#include "game/tables.h"
#include "game/team.h"
#include "game/racename.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"

/************************** WPlanetColonistTile **************************/

/** \class WPlanetColonistTile
    \brief Planet Colonists Tile

    Displays information about the planetary colony.

    <pre>
    Population:    XX
    Tax Rate:      XX%
    Colonists are XYZ.
    </pre> */

WPlanetColonistTile::WPlanetColonistTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(315, 3 * font_heights[FONT_NORMAL]), selection)
{
    setLayout(RightJustified, 0, true);
    addControl(GfxRect(0, 3 * font_heights[FONT_NORMAL] - 20, 20, 20), h.add(new UIButton("T", 't')));
}

void
WPlanetColonistTile::drawPlanetColonistTile(UIAbstractTableWriter& w)
{
    w.newLColumn();
    w.writeFixedCell(UIColor::tc_Static, _("Population:"));
    w.writeFixedCell(UIColor::tc_Static, _("Tax Rate:"));
    w.writeFixedCell(UIColor::tc_Static, _("Colonists are"));

    GPlanet* p = dynamic_cast<GPlanet*>(getCurrentObject());
    if (p != 0 && p->hasFullPlanetData()) {
        w.newLColumn(200);
        w.writeVariableCell(UIColor::tc_Green, clansToString(p->getColonists()));
        w.writeVariableCell(UIColor::tc_Green, format("%d%%", p->getColonistTax()));

        int happy = p->getColonistHappiness();
        UISkin::Color c = (happy <= 40
                           ? UIColor::tc_Red
                           : happy < 70
                           ? UIColor::tc_Yellow
                           : UIColor::tc_Green);
        w.writeVariableCell(c, format("%s (%d)", getHappinessName(happy), happy));
    }

    w.finish();
}

void
WPlanetColonistTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(5, 20), FONT_NORMAL);
    drawPlanetColonistTile(tw);
}



WNarrowPlanetColonistTile::WNarrowPlanetColonistTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(250, 3*font_heights[FONT_NORMAL]), selection)
{ }

void
WNarrowPlanetColonistTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());

    GPlanet* pl = dynamic_cast<GPlanet*>(getCurrentObject());
    if (!pl || !pl->isOwnerKnown()) {
        drawBackground(ctx, getInnerSize(0, 0));
        return;
    }

    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();

    int x = 0;
    int y = 0;
    int w = getInnerSize(0, 0).w;

    /* Anyone on this planet? */
    if (pl->getOwner() == 0) {
        /* No. */
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, _("No colonists."));
        return;
    }

    /* Normal display:
         Race, 10000 clans
         They like your leadership.
         Taxes: 99% (100 mc) */
    switch (getPlayerRelation(pl->getOwner())) {
     case is_Me:    ctx.setColor(UIColor::tc_Green);  break;
     case is_Enemy: ctx.setColor(UIColor::tc_Red);    break;
     case is_Ally:  ctx.setColor(UIColor::tc_Yellow); break;
     default:       ctx.setColor(UIColor::tc_Static); break;
    }
    if (pl->getColonists().isKnown()) {
        outTextF(ctx, x, y, w, format(_("%s, %d clan%!1{s%}"),
                                      player_racenames.getShortName(pl->getOwner()),
                                      numToString(pl->getColonists())));
    } else {
        outTextF(ctx, x, y, w, player_racenames.getShortName(pl->getOwner()));
    }
    y += font_heights[FONT_NORMAL];

    if (pl->getColonistHappiness().isKnown()) {
        int happy = pl->getColonistHappiness();
        ctx.setColor(happy <= 40
                     ? UIColor::tc_Red
                     : happy < 70
                     ? UIColor::tc_Yellow
                     : UIColor::tc_Green);
        outTextF(ctx, x, y, w, format("%s (%d)", getHappinessName(happy), happy));
    }
    y += font_heights[FONT_NORMAL];

    if (pl->getColonistTax().isKnown()) {
        const char* txt = _("Tax Rate:");
        const int wi = fonts[FONT_NORMAL]->getTextWidth(txt) + font_ems[FONT_NORMAL];
        ctx.setColor(UIColor::tc_Static);
        outText(ctx, x, y, txt);
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, x+wi, y, w-wi, format("%d%%", pl->getColonistTax()));
    }
}
