/**
  *  \file client/widgets/commandlist.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/widgets/commandlist.h"
#include "game/cmds.h"
#include "gfx/palette.h"
#include "ui/msgbox.h"
#include "util/misc.h"
#include "util/nls.h"
#include "ui/widgets/inputline.h"

/** Constructor. */
WCommandList::WCommandList(GCommandContainer& container)
    : UIListbox(0, 20, font_ems[FONT_NORMAL]*30, 0),
      container(container)
{
    update();
}

WCommandList::~WCommandList()
{ }

/** Update UIListbox parameters from container, and raise sig_moved. Used after
    all changes to the container object. */
void
WCommandList::update()
{
    setSize(std::distance(container.begin(), container.end()));
    adjustPosition();
    sig_moved.raise(getId());
}

/** Edit a command and add it to the container. This will add a new command or
    replace an existing one.
    \param s pre-initialize input field with this value. */
void
WCommandList::edit(string_t s)
{
    if (UIInputLine::run(s, 100, _("Auxiliary Commands"), _("Edit command:"), 30*font_ems[FONT_TITLE], UIInputLine::ilf_GameChars)) {
        GCommand* cmd = GCommand::parseCommand(s, true);
        if (cmd != 0) {
            const GCommand* addedCommand = container.setCommand(cmd);
            GCommandContainer::iterator it = std::find(container.begin(), container.end(), addedCommand);
            if (it != container.end())
                setCurrent(std::distance(container.begin(), it));

            update();

            drawWidget();
        } else {
            messageBox(_("This command was not recognized."), _("Auxiliary Commands"));
        }
    }
}

void
WCommandList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    GCommandContainer::iterator it = safeAdvance(container.begin(), container.end(), from + getTop());
    for (int line = from; line < to; ++line) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, line, &area);
        if (it != container.end()) {
            outTextF(ctx, area.x, area.y, area.w, " " + (*it)->getCommandText());
        }
        it = safeAdvance(it, container.end());
    }
}

bool
WCommandList::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && hasState(st_Focused)) {
        if (event.key.code == SDLK_DELETE) {
            if (safeAdvance(container.begin(), container.end(), getCurrent()) != container.end()) {
                if (UIMessageBox(_("Delete this command?"), _("Auxiliary Commands")).doYesNoDialog()) {
                    deleteCurrent();
                }
            }
            return true;
        }
        if (event.key.code == SDLK_INSERT) {
            editNew();
            return true;
        }
        if (event.key.code == ' ') {
            editCurrent();
            return true;
        }
    }
    return UIListbox::handleEvent(event, second_pass);
}

/** Delete current command. */
void
WCommandList::deleteCurrent()
{
    GCommandContainer::iterator it = safeAdvance(container.begin(), container.end(), getCurrent());
    if (it != container.end()) {
        container.removeCommand((*it)->getCommand(), (*it)->getId());
        update();
        drawWidget();
    }
}

/** Edit current command. */
void
WCommandList::editCurrent()
{
    if (const GCommand* cmd = getCurrentCommand()) {
        edit(cmd->getCommandText());
    }
}

/** Edit new command. */
void
WCommandList::editNew()
{
    edit(string_t());
}

/** Get current command.
    \return command, or 0 if none selected */
const GCommand*
WCommandList::getCurrentCommand()
{
    GCommandContainer::iterator it = safeAdvance(container.begin(), container.end(), getCurrent());
    if (it != container.end()) {
        return *it;
    } else {
        return 0;
    }
}
