/**
  *  \file client/widgets/objlist.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/objlist.h"
#include "client/widgets/helpwidget.h"
#include "ui/widgets/button.h"
#include "ui/layout.h"
#include "ui/window.h"
#include "cpluslib/holder.h"
#include "ui/skin.h"
#include "client/marks.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "client/chart.h"
#include "game/global.h"
#include "ui/widgets/quit.h"

WObjectList::WObjectList(int id, int lines, int width, GObjectList& list)
    : UIListbox(id, lines, width, list.size()),
      list(list),
      info_space(0),
      conn_turn_change(sig_turn_changed.add(this, &WObjectList::onTurnChange)),
      conn_universe_change(0)
{
    // Register universe observer
    onTurnChange();

    // In case first object is not selectable, bring cursor onto selectable object
    scrollTo(0, 1);
}

WObjectList::~WObjectList()
{ }

void
WObjectList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    for (int line = from; line < to; ++line) {
        const int index = getTop() + line;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, line, &area);
        if (index < int(list.size())) {
            GObjectReference ref = list[index];
            if (!ref.isValidInType() && dynamic_cast<GNoteObject*>(&*ref) == 0) {
                // Deleted
                ctx.setColor(UIColor::tc_Faded);
                ctx.useFont(*fonts[FONT_NORMAL]);
                area.consumeX(5);
                outTextF(ctx, area, _("(deleted)"));
            } else {
                UIColor::Color textcolor = UIColor::tc_Static;
                if (ref.isGameObject()) {
                    textcolor = getPlayerColor(ref->getOwner());
                }
                if (ref->getId() != 0) {
                    // A regular line
                    string_t info = getInfoForObject(ref);
                    if (ref->isMarked() && area.w > info_space+15) {
                        ctx.setColor(UIColor::tc_Selection);
                        drawSelection(ctx, area.splitX(15).getCenter(), 1, 2);
                    } else {
                        area.consumeX(5);
                    }

                    ctx.setColor(textcolor);
                    ctx.useFont(*fonts[FONT_NORMAL]);
                    if (info.length()) {
                        outTextF(ctx, area.splitX(area.w - info_space), ref->getName(GObject::DetailedName));
                        outTextF(ctx, area, info);
                    } else {
                        outTextF(ctx, area, ref->getName(GObject::DetailedName));
                    }
                } else {
                    // A divider
                    const int font = FONT_NORMAL_BOLD;
                    ctx.setColor(UIColor::tc_Faded);
                    ctx.useFont(*fonts[font]);

                    string_t name = ref->getName(GObject::DetailedName);
                    int y = area.y + font_heights[font]/2 - 1;
                    int max = std::min(fonts[font]->getTextWidth(name), area.w - 30);

                    drawHLine(ctx, area.x + 2, y, area.x + 28);
                    drawHLine(ctx, area.x + 32 + max, y, area.x + area.w - 2);
                    outTextF(ctx, area.x + 30, area.y, area.w - 30, name);
                }
            }
        }
    }
}

bool
WObjectList::isAccessible(int index)
{
    return index >= 0
        && index < int(list.size())
        && list[index].isValidInType()
        && list[index]->getId() != 0;
}

bool
WObjectList::selectObject(GObjectReference select)
{
    if (int pos = list.getIndexFor(select)) {
        scrollTo(pos-1, 1);
        return true;
    } else {
        return false;
    }
}

void
WObjectList::handleListChange(GObjectReference select)
{
    // ex onListChange

    // Update bounds
    setSize(list.size());

    // Go to object
    if (!selectObject(select)) {
        scrollTo(0, 1);
    }

    // Redraw everything
    drawWidget();
    sig_moved.raise(getId());
}

GObjectReference
WObjectList::getCurrentObject() const
{
    if (getCurrent() >= 0 && getCurrent() < int(list.size())) {
        return list[getCurrent()];
    } else {
        return GObjectReference();
    }
}

GObjectReference
WObjectList::doStandardDialog2(string_t title, string_t help_id)
{
    // Can we do something?
    if (list.size() == 0) {
        return GObjectReference();
    }

    // Okay, go ahead.
    ObjectHolder holder;

    UIWindow win(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(*this);

    UIGroup bgroup(UIHBoxLayout::instance5);
    bgroup.add(holder.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    bgroup.add(holder.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    bgroup.add(holder.add(new UISpacer()));
    if (!help_id.empty()) {
        bgroup.add(holder.add(new UIButton(_("Help"), 'h')));
        bgroup.add(holder.add(new WHelpWidget(help_id)));
    }
    win.add(bgroup);
    win.add(holder.add(new UIQuit(0)));
    win.setFocusedWidget(*this);
    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    if (rv) {
        return getCurrentObject();
    } else { 
        return GObjectReference();
    }
}

/** Set room for additional information. */
void
WObjectList::setInfoSpace(int space)
{
    this->info_space = space;
}

/** Fetch additional information for an object. Derived classes can override
    this to provide additional information for an object. For example,
    in proxy transfer, PCC will show current ship/ship transfers when
    selecting the proxy. */
string_t
WObjectList::getInfoForObject(GObjectReference /*ref*/)
{
    return string_t();
}

/** Signal handler: turn change. Hook into new universe and redraw. */
void
WObjectList::onTurnChange()
{
    conn_universe_change = getDisplayedTurn().getCurrentUniverse().sig_pre_update.add(this, &WObjectList::onUniverseChange);
    drawWidget(false);
}

/** Signal handler: universe possibly changed. */
void
WObjectList::onUniverseChange()
{
    bool changed_any = false;
    bool changed_current = false;

    /* Check for changes */
    int index = getTop();
    for (int line = 0; line <= getPageSize(); ++line) {
        /* "line <= getPageSize()" to allow for partial last line */
        if (index < int(list.size()) && (!list[index].isValidInType() || list[index]->isDirty())) {
            if (index == getCurrent()) {
                changed_current = true;
            } else {
                changed_any = true;
            }
        }
        ++index;
    }

    /* Apply changes */
    if (changed_any) {
        /* some objects changed */
        drawWidget(false);
    } else if (changed_current) {
        /* current object changed, we have a special update mechanism for that */
        updateCurrent();
    } else {
        /* no change */
    }
}
