/**
  *  \file game/actions/changebuildqueue.cc
  *  \brief Class GChangeBuildQueue
  *
  *  Build rules:
  *
  *  Host:
  *    - PBx friendly codes are not case-sensitive
  *    - identical PBx friendly codes ok, will build both
  *    - can priority-build more than 10 ships
  *
  *  PHost PAL:
  *    - PBx friendly codes are case-sensitive
  *    - only one PBx per slot, reorders builds
  *    - clones in build queue
  *
  *  PHost PBP:
  *    - PBx friendly codes are case-sensitive
  *    - only one PBx per slot
  *    - clones in build queue
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/actions/changebuildqueue.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/fcode.h"
#include "int/exec.h"
#include "int/if/basepred.h"
#include "int/autotask.h"

namespace {
    const int MAX_PRIORITY_VALUE = 9;

    /* Check priority friendly code. */
    int checkPriorityCode(const string_t& str, const GHost& host)
    {
        if (str.size() == 3
            && (str[0] == 'P' || (!host.isPHost() && str[0] == 'p'))
            && (str[1] == 'B' || (!host.isPHost() && str[1] == 'b'))
            && (str[2] >= '1' && str[2] <= '9'))
        {
            return (str[2] - '0');
        } else {
            return 0;
        }
    }

    string_t makePriorityCode(int pri)
    {
        return string_t("PB") + char('0' + pri);
    }


    int findCloningShipAt(GUniverse& univ, GPoint pt)
    {
        GPlayedShipType& ships = univ.ty_played_ships;
        for (int i = ships.findNextIndex(0); i != 0; i = ships.findNextIndex(i)) {
            GShip& sh = univ.getShip(i);
            if (sh.isPlayable(GObject::ReadOnly) && sh.getPos() == pt && sh.getFCode() == "cln") {
                return i;
            }
        }
        return 0;
    }

    void canonicalizeBuildOrder(TBuildOrder& o)
    {
        if (!o.torp_count)
            o.torp_type = 0;
        if (!o.beam_count)
            o.beam_type = 0;
        o.unused = 0;
    }

    bool isEqual(TBuildOrder a, TBuildOrder b)
    {
        canonicalizeBuildOrder(a);
        canonicalizeBuildOrder(b);
        return a.hull == b.hull
            && a.engine == b.engine
            && a.beam_type == b.beam_type
            && a.beam_count == b.beam_count
            && a.torp_type == b.torp_type
            && a.torp_count == b.torp_count;
    }
}

/*
 *  Sort Predicate
 *
 *  This tries to sort the build orders into the order in which they will be executed.
 */
class GChangeBuildQueue::Sorter {
 public:
    Sorter(const GChangeBuildQueue& parent)
        : m_parent(parent)
        { }
    bool operator()(const LocalInfo& a, const LocalInfo& b) const
        {
            // Planned goes last
            bool planA = (a.plannedHullId != 0);
            bool planB = (b.plannedHullId != 0);
            if (planA != planB) {
                return planA < planB;
            }

            // Check priorities only for not planned
            if (!planA) {
                // Parse friendly codes
                int valA = checkPriorityCode(a.friendlyCode, m_parent.m_host);
                int valB = checkPriorityCode(b.friendlyCode, m_parent.m_host);

                // Prioritized goes before unprioritized
                bool priA = (valA != 0);
                bool priB = (valB != 0);
                if (priA != priB) {
                    return priA > priB;
                }

                if (priA) {
                    // Two priorized orders: lower values go first
                    if (valA != valB) {
                        return valA < valB;
                    }
                } else {
                    // Unpriorized: use queue order if known
                    int knownA = a.queuePosition != 0;
                    int knownB = b.queuePosition != 0;
                    if (knownA != knownB) {
                        return knownA > knownB;
                    }
                    if (a.queuePosition != b.queuePosition) {
                        return a.queuePosition < b.queuePosition;
                    }
                }
            }

            // Use Id as tie-breaker
            return a.planetId < b.planetId;
        }
 private:
    const GChangeBuildQueue& m_parent;
};

// Constructor.
GChangeBuildQueue::GChangeBuildQueue(GGameTurn& turn,
                                     GHost host,
                                     const GGameConfig& config,
                                     RandomNumberGenerator& rng,
                                     int viewpointPlayer)
    : m_turn(turn),
      m_host(host),
      m_config(config),
      m_availablePoints()
{
    init(rng, viewpointPlayer);
}

// Destructor.
GChangeBuildQueue::~GChangeBuildQueue()
{ }

void
GChangeBuildQueue::addPlannedBuilds(const IntProcessList& list)
{
    for (size_t i = 0, n = list.size(); i < n; ++i) {
        // Check process state: only suspended processes.
        // For PCC2, this means we won't see the current base if this function is called from the base task screen.
        // In that case, the content of the IntExecutionContext may be different from what the user sees, anyway.
        IntExecutionContext& p = *list[i];
        if (p.getProcessKind() == IntExecutionContext::pkBaseTask && p.getState() == IntExecutionContext::psSuspended) {
            if (const GPlanet* pl = dynamic_cast<const GPlanet*>(p.getInvokingObject())) {
                if (pl->isPlayable(GObject::Playable) && !hasPlanet(pl->getId())) {
                    // Load from process. We don't plan on ever saving.
                    IntAutoTaskEditor ed;
                    ed.loadFromProcess(p);

                    // Parse
                    WBaseTaskBuildCommandParser cmd;
                    cmd.predictStatement(ed, ed.getPC());

                    // Build command?
                    int hullType = cmd.getOrder().hull;
                    if (hullType > 0 && hullType <= getNumHulls()) {
                        m_info.push_back(LocalInfo(pl->getId(), 0, hullType, pl->getFCode(), pl->getFCode(), 0, true));
                    }
                }
            }
        }
    }
    sort();
}

// Set available build points.
void
GChangeBuildQueue::setAvailableBuildPoints(mn32_t points)
{
    m_availablePoints = points;
}

// Prepare data into output format.
void
GChangeBuildQueue::describe(Infos_t& result) const
{
    bool pointsOK = m_availablePoints.isKnown();
    int32_t points = pointsOK ? m_availablePoints.getValue() : 0;
    for (size_t i = 0, n = m_info.size(); i < n; ++i) {
        // Input
        const LocalInfo& in = m_info[i];
        const GPlanet& pl = m_turn.getCurrentUniverse().getPlanet(in.planetId);
        int player = pl.getOwner();

        // Make empty slot
        result.push_back(Info());
        Info& out = result.back();

        // Simple content
        out.planetId = in.planetId;
        out.planetName = pl.getName(GObject::PlainName);
        out.friendlyCode = in.friendlyCode;
        out.queuePosition = in.queuePosition;
        out.conflict = (i > 0 && m_info[i-1].friendlyCode == out.friendlyCode && checkPriorityCode(out.friendlyCode, m_host) != 0);
        out.hasPriority = checkPriorityCode(in.friendlyCode, m_host);
        out.playable = in.playable;
        out.planned = (in.plannedHullId != 0);
        out.isChange = (in.friendlyCode != pl.getFCode());

        // Ship being built
        int hullMass = 0;
        bool isClone = false;
        if (in.plannedHullId != 0) {
            const GHull& hull = getHull(in.plannedHullId);
            out.hullName = hull.getName();
            out.hullNr = hull.getId();
            out.actionName = format(_("Plan %s"), out.hullName);
            out.action = PlanShip;
            hullMass = hull.getMass();
        } else if (in.cloningShipId == 0) {
            const GHull& hull = getHull(pl.getBaseBuildHull());
            out.hullName = hull.getName();
            out.hullNr = hull.getId();
            out.actionName = format(_("Build %s"), out.hullName);
            out.action = BuildShip;
            hullMass = hull.getMass();
        } else {
            const GShip& sh = m_turn.getCurrentUniverse().getShip(in.cloningShipId);
            string_t shipName = sh.getName(GObject::PlainName);
            if (shipName.empty()) {
                shipName = format(_("Ship #%d"), sh.getId());
            }
            out.actionName = format(_("Clone %s"), shipName);
            out.action = CloneShip;
            out.hullName = sh.getHull().getName();
            out.hullNr = sh.getHullId();

            hullMass = sh.getHull().getMass();
            isClone = true;
        }

        // Points
        if (m_host.isPBPGame(/*m_config*/) && hullMass != 0) {
            int32_t pointsRequired;
            if (m_host.isPHost()) {
                pointsRequired = std::max(hullMass * m_config.PBPCostPer100KT(player) / 100,
                                          m_config.PBPMinimumCost(player));
                if (isClone) {
                    pointsRequired = pointsRequired * m_config.PBPCloneCostRate(player) / 100;
                }
            } else {
                pointsRequired = (hullMass + 49) / 50;
            }
            out.pointsRequired = mn32_t(pointsRequired);

            if (pointsOK) {
                out.pointsAvailable = mn32_t(points);
                points = std::max(int32_t(0), points - pointsRequired);
            }
        }
    }
}

// Set priority of a build order.
void
GChangeBuildQueue::setPriority(size_t slot, int pri)
{
    if (slot < m_info.size()) {
        LocalInfo& info = m_info[slot];
        if (info.playable && info.plannedHullId == 0) {
            if (pri >= 1 && pri <= MAX_PRIORITY_VALUE) {
                info.friendlyCode = makePriorityCode(pri);
            } else {
                info.friendlyCode = info.oldFriendlyCode;
            }
            sort();
        }
    }
}

// Increase a slot's priority (build earlier).
void
GChangeBuildQueue::increasePriority(size_t slot)
{
    const size_t n = m_info.size();
    if (slot < n) {
        LocalInfo& info = m_info[slot];
        if (info.playable && info.plannedHullId == 0) {
            const int current = checkPriorityCode(info.friendlyCode, m_host);
            if (current == 0) {
                // Unpriorized to lowest priority: set to one-plus-highest used value
                // (need to check items before slot only)
                // If no priority codes are currently used, this conveniently sets the value to PB1.
                int maxValue = 0;
                for (size_t i = 0; i < slot; ++i) {
                    maxValue = std::max(maxValue, checkPriorityCode(m_info[i].friendlyCode, m_host));
                }
                info.friendlyCode = makePriorityCode(std::min(MAX_PRIORITY_VALUE, 1 + maxValue));
            } else if (current > 1) {
                // Higher priority (lower value).
                // Set priority value. If this causes a duplicate, try to move these out of the way.
                info.friendlyCode = makePriorityCode(current-1);
                avoid(current-1, current, slot);
            } else {
                // Cannot go higher
            }
            sort();
        }
    }
}

// Decrease a slot's priority (build later).
void
GChangeBuildQueue::decreasePriority(size_t slot)
{
    const size_t n = m_info.size();
    if (slot < n) {
        LocalInfo& info = m_info[slot];
        if (info.playable && info.plannedHullId == 0) {
            const int current = checkPriorityCode(info.friendlyCode, m_host);
            if (current == 0) {
                // Cannot go any lower
            } else if (current == 9 || (slot+1 < m_info.size() && checkPriorityCode(m_info[slot+1].friendlyCode, m_host) == 0)) {
                // Least priority to unprioritized
                info.friendlyCode = info.oldFriendlyCode;
            } else {
                // Lower priority (higher value)
                info.friendlyCode = makePriorityCode(current+1);
                avoid(current+1, current, slot);
            }
            sort();
        }
    }
}

// Write all changes back to universe.
void
GChangeBuildQueue::commit()
{
    for (size_t i = 0, n = m_info.size(); i < n; ++i) {
        GPlanet& pl = m_turn.getCurrentUniverse().getPlanet(m_info[i].planetId);
        if (pl.isPlayable(GObject::Playable) && m_info[i].plannedHullId == 0) {
            pl.setFCode(m_info[i].friendlyCode);
        }
    }
}

void
GChangeBuildQueue::init(RandomNumberGenerator& /*rng*/, int viewpointPlayer)
{
    GUniverse& univ = m_turn.getCurrentUniverse();
    GPlayedBaseType& bases = univ.ty_played_bases;
    for (int i = bases.findNextIndex(0); i != 0; i = bases.findNextIndex(i)) {
        GPlanet& p = univ.getPlanet(i);
        int owner = p.getOwner();
        GPoint pt = p.getPos();
        if (owner == viewpointPlayer) {
            // Existing friendly code
            string_t friendlyCode = p.getFCode();

            // Old (revert) friendly code
            string_t oldFriendlyCode = checkPriorityCode(friendlyCode, m_host) == 0
                ? friendlyCode
                : GFCode::generateRandomFCode();

            // Find ship being cloned here. Only for PHost; cloning has entirely different queue in Host.
            int cloningShipId = 0;
            int previouslyCloningShipId = 0;
            if (m_host.isPHost() && m_config.AllowShipCloning()) {
                cloningShipId = findCloningShipAt(univ, pt);
                previouslyCloningShipId = findCloningShipAt(m_turn.getPreviousUniverse(), pt);
            }

            // Ship being built here
            int buildingHullId = p.getBaseBuildHull();
            if (cloningShipId != 0 || buildingHullId != 0) {
                // New build order
                bool isNew = false;
                if (m_host.isPHost()) {
                    // For PHost, check whether order was changed
                    if (cloningShipId != 0) {
                        // Cloning ship: ID must not have changed (build order not relevant)
                        if (previouslyCloningShipId != cloningShipId) {
                            isNew = true;
                        }
                    } else {
                        // Not cloning: build order must be the same, no clone
                        if (!isEqual(m_turn.getPreviousUniverse().getPlanet(i).getBaseBuildOrder(), p.getBaseBuildOrder())
                            || previouslyCloningShipId != 0)
                        {
                            isNew = true;
                        }
                    }
                }

                // Queue position
                int queuePosition = isNew ? 0 : p.getBaseQueuePosition().isKnown() ? p.getBaseQueuePosition() : 0;

                // Playability
                bool playable = p.isPlayable(GObject::Playable);

                // Remember it
                m_info.push_back(LocalInfo(i, cloningShipId, 0, friendlyCode, oldFriendlyCode, queuePosition, playable));
            }
        }
    }
    sort();
}

bool
GChangeBuildQueue::hasPlanet(int id) const
{
    for (size_t i = 0, n = m_info.size(); i < n; ++i) {
        if (m_info[i].planetId == id) {
            return true;
        }
    }
    return false;
}

void
GChangeBuildQueue::sort()
{
    std::sort(m_info.begin(), m_info.end(), Sorter(*this));
}

void
GChangeBuildQueue::avoid(int setThis, int toThis, size_t slot)
{
    for (size_t i = 0, n = m_info.size(); i < n; ++i) {
        if (i != slot && m_info[i].playable && checkPriorityCode(m_info[i].friendlyCode, m_host) == setThis) {
            m_info[i].friendlyCode = makePriorityCode(toThis);
        }
    }
}
