/**
  *  \file game/actions/changebuildqueue.h
  *  \brief Class GChangeBuildQueue
  *
  *  (ported from c2ng game::actions::ChangeBuildQueue)
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_ACTIONS_CHANGEBUILDQUEUE_H
#define PCC_V2_GAME_ACTIONS_CHANGEBUILDQUEUE_H

#include <vector>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "io/io.h"              // mn32_t (LongProperty_t)
#include "game/global.h"
#include "game/host.h"
#include "game/config.h"
#include "util/random.h"
#include "int/process.h"

/** Changing build queue priorities.
    Provides a simple view of the build queue,
    and allows changing priorities (distributing PBx fcodes).

    - create object
    - if desired, call setAvailableBuildPoints()
    - call increasePriority(), decreasePriority(), setPriority() as needed
    - call commit() to write back */
class GChangeBuildQueue {
 public:
    /** Action. */
    enum Action {
        BuildShip,                          ///< Building a ship.
        CloneShip,                          ///< Cloning a ship.
        PlanShip                            ///< Planning a ship (waiting for resources).
    };

    /** Information about a build order in prepared format. */
    struct Info {
        int planetId;                       ///< Planet Id.
        Action action;                      ///< Action type.
        string_t planetName;                ///< Planet name.
        string_t actionName;                ///< Human-readable description of build order.
        string_t friendlyCode;              ///< Friendly code.
        string_t hullName;                  ///< Hull name.
        int hullNr;                         ///< Hull number.
        int queuePosition;                  ///< Queue position (0 if not known or new order).
        mn32_t pointsRequired;              ///< Number of build points required to build this ship.
        mn32_t pointsAvailable;             ///< Number of build points available for this ship.
        bool hasPriority;                   ///< true if this build order has a priority FC.
        bool conflict;                      ///< true if this priority order conflicts with others (same FC).
        bool playable;                      ///< true if this slot can be modified.
        bool planned;                       ///< true if this is a planned build.
        bool isChange;                      ///< true if this item is a change to the game.

        Info()
            : planetId(), action(), planetName(), actionName(),
              friendlyCode(), hullName(), hullNr(),
              queuePosition(), pointsRequired(), pointsAvailable(),
              hasPriority(), conflict(), playable(), planned(), isChange()
            { }
    };

    /** Information about all build orders in prepared format. */
    typedef std::vector<Info> Infos_t;


    /** Constructor.
        \param turn Turn (containing current and previous universe)
        \param host Host version
        \param config Host configuration
        \param rng Random number generator (required to generate fall-back friendly codes)
        \param viewpointPlayer Handle this player's build orders */
    explicit GChangeBuildQueue(GGameTurn& turn,
                              GHost host,
                              const GGameConfig& config,
                              RandomNumberGenerator& rng,
                              int viewpointPlayer);

    /** Destructor. */
    ~GChangeBuildQueue();

    /** Add planned build orders from a process list.
        \param list Process list */
    void addPlannedBuilds(const IntProcessList& list);

    /** Set available build points.
        Call this to populate the "pointsAvailable" field.
        \param points Build points */
    void setAvailableBuildPoints(mn32_t points);

    /** Prepare data into output format.
        Data is reported sorted, highest priority first.
        \param result [out] Data */
    void describe(Infos_t& result) const;

    /** Set priority of a build order.
        \param slot Slot number
        \param pri New priority (1-9 = PBx, 0 = no priority order) */
    void setPriority(size_t slot, int pri);

    /** Increase a slot's priority (build earlier).
        \param slot Slot number */
    void increasePriority(size_t slot);

    /** Decrease a slot's priority (build later).
        \param slot Slot number */
    void decreasePriority(size_t slot);

    /** Write all changes back to universe. */
    void commit();

 private:
    GGameTurn& m_turn;
    // const game::spec::ShipList& m_shipList;
    GHost m_host;
    const GGameConfig& m_config;

    void init(RandomNumberGenerator& rng, int viewpointPlayer);

    bool hasPlanet(int id) const;
    void sort();
    void avoid(int setThis, int toThis, size_t slot);

    struct LocalInfo {
        int planetId;               ///< Planet Id.
        int cloningShipId;          ///< Cloning ship Id.
        int plannedHullId;          ///< Planned hull Id (for auto task).
        string_t friendlyCode;      ///< Current friendly code.
        string_t oldFriendlyCode;   ///< Friendly code to revert to (old or random).
        int queuePosition;          ///< Current queue position if known. 0 for new orders.
        bool playable;

        LocalInfo(int planetId, int cloningShipId, int plannedHullId, string_t friendlyCode, string_t oldFriendlyCode, int queuePosition, bool playable)
            : planetId(planetId), cloningShipId(cloningShipId), plannedHullId(plannedHullId), friendlyCode(friendlyCode), oldFriendlyCode(oldFriendlyCode), queuePosition(queuePosition), playable(playable)
            { }
    };
    class Sorter;

    std::vector<LocalInfo> m_info;
    mn32_t m_availablePoints;
};

#endif
