/**
  *  \file game/objlist.h
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJLIST_H
#define PCC_V2_GAME_OBJLIST_H

#include <vector>
#include "cpluslib/types.h"
#include "game/objtypes.h"
#include "game/obj.h"
#include "game/objnote.h"

/** Flags for ship list and GObjectList::addObjectsAt(). */
enum {
    /* addObjectsAt */
    ls_IncludeForeignShips = 1,
    ls_IncludePlanet       = 2,
    ls_SafeShipsOnly       = 4,

    /* ship list */
    ls_AllowForeignShips   = 8,
    ls_AllowPlanet         = 16,
    ls_EarlyExit           = 32
};

/** Sort orders. */
enum {
    SortById          = 0,
    SortByOwner       = 1,
    SortByHull        = 2,
    SortByMass        = 3,
    SortByFleet       = 4,
    SortByTowGroup    = 5,
    SortByBattleOrder = 6,
    SortByLocation    = 7,
    SortByHullMass    = 8,
    SortByDamage      = 9,
    SortByName        = 10,
    SortByNewPosition = 11
};

class GObjectReference {
 public:
    GObjectReference();
    GObjectReference(GObjectType& type, int index);

    // Use default operator==, copy constructor

    bool         isValid() const;
    bool         isValidInType() const;
    GObjectType& getObjectType() const;
    GUniverse*   getUniverse() const;
    int          getObjectIndex() const;
    bool         isGameObject() const;

    GObject& operator*() const;
    GObject* operator->() const;

    bool operator==(const GObjectReference& o) const;
    bool operator!=(const GObjectReference& o) const;

 private:
    GObjectType* type;
    int index;
};

class GObjectDivider {
 public:
    virtual ~GObjectDivider() { }
    virtual string_t getDivider(GObject& obj) = 0;
};

class GObjectCompare {
 public:
    virtual ~GObjectCompare() { }
    virtual int compare(GObject& lhs, GObject& rhs) = 0;
};

/** Object list. Container for a list of objects (GObjectReference's). It offers a number
    of functions to add objects and bring them into a form convenient to the user.
    It also contains a GNoteContainer to allow adding additional selectable items
    and dividers. */
class GObjectList : public GObjectType {
 public:
    typedef std::vector<GObjectReference> container_t;
    typedef container_t::const_iterator iterator;
    typedef container_t::size_type size_type;

    // Construction:
    GObjectList();
    ~GObjectList();

    // Adding things:
    void         addObject(GObjectReference ref);
    void         addObject(GObjectType& type, int index);
    void         addObjectsAt(GUniverse& univ, GPoint pt, int flags, int exclude_ship);
    void         addNonObject(int id, string_t name);

    // Manipulation:
    void         sort(int func(GObject&,GObject&));
    void         sort(int32_t func(GObject&));
    void         sort(GObjectCompare& cmp);
    void         clear();
    void         addDividers(string_t func(GObject&));
    void         addDividers(GObjectDivider& func);
    void         removeDividers();
    void         postprocessUser(bool with_dividers);

    // Container access: (0-based)
    iterator     begin() const;
    iterator     end() const;
    size_type    size() const;
    const GObjectReference& operator[](size_type index) const;

    // GObjectType: (1-based!)
    GObject&     getObjectByIndex(int index);
    GUniverse*   getUniverseByIndex(int index);
    bool         isValidIndex(int index) const;
    int          getNextIndex(int index) const;
    int          getPreviousIndex(int index) const;
    const GObjectReference& getObjectReferenceByIndex(int index) const;

    int          getIndexFor(GObjectReference id) const;
    int          getIndexFor(const GObject& obj) const;

    size_type    countGameObjects() const;

 private:
    container_t    list;
    GNoteContainer notes;
};

int32 sortById(GObject& o);
int32 sortByOwner(GObject& o);
int   sortByName(GObject& l, GObject& r);
int32 sortByHull(GObject& o);
int32 sortByMass(GObject& o);
int32 sortByHullMass(GObject& o);
int32 sortByDamage(GObject& o);
int32 sortByFleet(GObject& o);
int32 sortByTowGroup(GObject& o);
int   sortByBattleOrder(GObject& o);
int   sortByLocation(GObject& l, GObject& r);
int   sortByNewLocation(GObject& l, GObject& r);

string_t diviOwner(GObject& o);
string_t diviHull(GObject& o);
string_t diviFleet(GObject& o);
string_t diviTowGroup(GObject& o);
string_t diviBattleOrder(GObject& o);
string_t diviLocation(GObject& o);
string_t diviNewLocation(GObject& o);

int  getSortOrderConfig();
void setSortOrderConfig(int n);

/**************************** Inline Functions ***************************/

/** Check validity.
    Check whether this GObjectReference refers to an object.
    The object may not actually be valid. */
inline bool
GObjectReference::isValid() const
{
    return type != 0;
}

/** Check validity in type.
    Checks whether this GObjectReference refers to an object which its
    containing type considers valid. */
inline bool
GObjectReference::isValidInType() const
{
    return type != 0
        && type->isValidIndex(index);
}

inline GObjectType&
GObjectReference::getObjectType() const
{
    return *type;
}

inline GUniverse*
GObjectReference::getUniverse() const
{
    return type->getUniverseByIndex(index);
}

inline int
GObjectReference::getObjectIndex() const
{
    return index;
}

inline bool
GObjectReference::isGameObject() const
{
    return type != 0
        && type->isValidIndex(index)
        && dynamic_cast<GNoteContainer*>(type) == 0;
}

inline GObject&
GObjectReference::operator*() const
{
    return type->getObjectByIndex(index);
}

inline GObject*
GObjectReference::operator->() const
{
    return &type->getObjectByIndex(index);
}

inline bool
GObjectReference::operator==(const GObjectReference& o) const
{
    return type == o.type
        && index == o.index;
}

inline bool
GObjectReference::operator!=(const GObjectReference& o) const
{
    return type != o.type
        || index != o.index;
}

#endif
