/**
  *  \file game/palliance.cc
  *  \brief PHost alliance handling
  *
  *  The integration of this module is sub-optimal.
  *  We create all necessary data from the util.dat parser,
  *  including the main data which only depends on the host version.
  *  Normally, we should know the host version in advance and create everything.
  *  This would also allow proper "if (host.XXX)" tests instead of hardcoded conditionals.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "game/palliance.h"
#include "game/config.h"
#include "util/nls.h"
#include "game/cmds.h"
#include "util/misc.h"

/* Main */
static const char MAIN_ID[] = "phost.ally";

static const uint8_t MAIN_FLAG = 0x20;

/* Levels */
static const size_t NUM_LEVELS = 5;
static const size_t COMBAT_LEVEL_INDEX = 3;

static const char*const LEVEL_IDS[NUM_LEVELS] = {
    "phost.s",
    "phost.p",
    "phost.m",
    "phost.c",
    "phost.v",
};

static const char*const LEVEL_NAMES[NUM_LEVELS] = {
    N_("Ship alliance"),
    N_("Planet alliance"),
    N_("Minefield alliance"),
    N_("Combat alliance"),
    N_("Vision alliance"),
};

static const uint8_t LEVEL_FLAGS[NUM_LEVELS] = {
    0x01,
    0x02,
    0x04,
    0x08,
    0x10,
};

static const char LEVEL_LETTERS[NUM_LEVELS] = { 's', 'p', 'm', 'c', 'v' };

/* Enemies */
static const char ENEMY_ID[] = "phost.enemy";

/******************************* Utilities *******************************/

/** Check for valid player number. */
static inline bool
isValidPlayer(int nr)
{
    return GPlayerArray<GAllianceOffer::OfferType>::isValidIndex(nr);
}

/** Convert "add"/"drop" to an offer type.
    Used for "allies" and "enemies". */
static GAllianceOffer::OfferType
convertFromAddDrop(const string_t& name)
{
    using std::tolower;
    if (name.size() > 0 && tolower(name[0]) == 'a') {
        return GAllianceOffer::Yes;
    } else {
        return GAllianceOffer::No;
    }
}

/** Conver offer type to "add"/"drop". */
static inline const char*
convertToAddDrop(GAllianceOffer::OfferType type)
{
    if (type == GAllianceOffer::Yes || type == GAllianceOffer::Conditional) {
        return "add";
    } else {
        return "drop";
    }
}

/** Process level offer ("+p") into an alliance object.
    \param allies Alliance object
    \param player Target player
    \param word word ("+p") from command */
static void
processLevelOffer(GAlliances& allies, int player, const string_t& word)
{
    // Must at least have a control character and a level letter
    if (word.size() < 2) {
        return;
    }

    // Determine mode
    GAllianceOffer::OfferType type;
    switch (word[0]) {
     case '+': type = GAllianceOffer::Yes;         break;
     case '-': type = GAllianceOffer::No;          break;
     case '~': type = GAllianceOffer::Conditional; break;
     default:  return;
    }

    // Determine level
    using std::tolower;
    for (size_t i = 0; i < NUM_LEVELS; ++i) {
        if (LEVEL_LETTERS[i] == tolower(word[1])) {
            if (GAllianceOffer* pOffer = allies.getMutableOffer(allies.find(LEVEL_IDS[i]))) {
                pOffer->newOffer[player] = type;
            }
            break;
        }
    }
}

/** Clear all new offers for a given offer Id.
    \param allies Alliance object
    \param id Identifier (MAIN_ID etc.) */
static void
clearAll(GAlliances& allies, const char* id)
{
    if (GAllianceOffer* pOffer = allies.getMutableOffer(allies.find(id))) {
        pOffer->newOffer = pOffer->oldOffer;
    }
}

/************************* GPHostAllianceHandler *************************/

/** Constructor. */
GPHostAllianceHandler::GPHostAllianceHandler(GCommandContainer& commands)
    : commands(commands)
{ }

/** Postprocess after game loading: parse commands into internal state. */
void
GPHostAllianceHandler::postprocess(GAlliances& allies)
{
    // Clear everything
    clearAll(allies, MAIN_ID);
    for (size_t i = 0; i < NUM_LEVELS; ++i) {
        clearAll(allies, LEVEL_IDS[i]);
    }
    clearAll(allies, ENEMY_ID);

    // Parse commands
    using std::tolower;
    for (GCommandContainer::const_iterator it = commands.begin(), e = commands.end(); it != e; ++it) {
        const GCommand& cmd = **it;
        switch (cmd.getCommand()) {
         case GCommand::phc_AddDropAlly:
            // Id = player, Arg = "add" or "drop"
            if (GAllianceOffer* pOffer = allies.getMutableOffer(allies.find(MAIN_ID))) {
                if (isValidPlayer(cmd.getId())) {
                    pOffer->newOffer[cmd.getId()] = convertFromAddDrop(cmd.getArg());
                }
            }
            break;

         case GCommand::phc_ConfigAlly:
            // Id = player, Arg = "+c -s ~m"
            if (isValidPlayer(cmd.getId())) {
                string_t arg = cmd.getArg();
                do {
                    processLevelOffer(allies, cmd.getId(), strFirst(arg, " "));
                } while (strRemove(arg, " "));
            }
            break;

         case GCommand::phc_Enemies:
            // Id = player, Arg = "add" or "drop"
            if (GAllianceOffer* pOffer = allies.getMutableOffer(allies.find(ENEMY_ID))) {
                if (isValidPlayer(cmd.getId())) {
                    pOffer->newOffer[cmd.getId()] = convertFromAddDrop(cmd.getArg());
                }
            }
            break;

         default:
            break;
        }
    }
}

/** Process changes to alliance object: generate command messages. */
void
GPHostAllianceHandler::handleChanges(const GAlliances& allies)
{
    if (const GAllianceOffer* pMainOffer = allies.getOffer(allies.find(MAIN_ID))) {
        // Allies
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            // Transmit main offer
            bool sendLevels = false;
            if (pMainOffer->oldOffer[i] != pMainOffer->newOffer[i]) {
                commands.setCommand(GCommand::phc_AddDropAlly, i, convertToAddDrop(pMainOffer->newOffer[i]));
                sendLevels = true;
            } else {
                commands.removeCommand(GCommand::phc_AddDropAlly, i);
            }

            // Transmit levels. We always send a complete level list if anything changes.
            // We also send a complete list if the main offer changed.
            // This is the same which PCC 1.x does; it offers some robustness against things getting out of sync.
            string_t levelStr;
            for (size_t lvl = 0; lvl < NUM_LEVELS; ++lvl) {
                if (const GAllianceOffer* pLevelOffer = allies.getOffer(allies.find(LEVEL_IDS[lvl]))) {
                    // Register changes
                    if (pLevelOffer->oldOffer[i] != pLevelOffer->newOffer[i]
                        && pLevelOffer->newOffer[i] != GAllianceOffer::Unknown)
                    {
                        sendLevels = true;
                    }

                    // Build command
                    if (!levelStr.empty()) {
                        levelStr += ' ';
                    }
                    switch (pLevelOffer->newOffer[i]) {
                     case GAllianceOffer::Yes:         levelStr += '+'; break;
                     case GAllianceOffer::No:          levelStr += '-'; break;
                     case GAllianceOffer::Conditional: levelStr += '~'; break;
                     case GAllianceOffer::Unknown:     levelStr += '-'; break;
                    }
                    levelStr += LEVEL_LETTERS[lvl];
                }
            }
            if (sendLevels) {
                commands.setCommand(GCommand::phc_ConfigAlly, i, levelStr);
            } else {
                commands.removeCommand(GCommand::phc_ConfigAlly, i);
            }
        }
    }

    if (const GAllianceOffer* pEnemies = allies.getOffer(allies.find(ENEMY_ID))) {
        // Enemies
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            if (pEnemies->oldOffer[i] != pEnemies->newOffer[i]) {
                commands.setCommand(GCommand::phc_Enemies, i, convertToAddDrop(pEnemies->newOffer[i]));
            } else {
                commands.removeCommand(GCommand::phc_Enemies, i);
            }
        }
    }
}

/** Process a version record from util.dat.
    Because this always is the first thing in util.dat, we can use it to determine which levels we have,
    and create them.
    \param cmd Command container
    \param alliances Alliance container
    \param version Version number from util.dat. */
void
GPHostAllianceHandler::processVersion(GCommandContainer& cmd, GAlliances& alliances, uint32_t version)
{
    bool added = false;
    if (config.CPEnableAllies() && alliances.find(MAIN_ID) == GAlliances::nil) {
        // We have not yet done our work; do it now:
        // add the main level
        alliances.addLevel(GAllianceLevel(_("Alliance offer"), MAIN_ID, GAllianceLevel::IsOffer));

        // add the sub levels
        for (size_t i = 0; i < NUM_LEVELS; ++i) {
            int flags = GAllianceLevel::NeedsOffer | GAllianceLevel::AllowConditional;
            if (i == COMBAT_LEVEL_INDEX) {
                flags |= GAllianceLevel::IsCombat;
            }
            alliances.addLevel(GAllianceLevel(_(LEVEL_NAMES[i]), LEVEL_IDS[i], flags));
        }
        added = true;
    }

    if (version >= 400008 && config.CPEnableEnemies() && alliances.find(ENEMY_ID) == GAlliances::nil) {
        // add the enemies, if supported by host
        alliances.addLevel(GAllianceLevel(_("Enemy"), ENEMY_ID, GAllianceLevel::IsEnemy));
        added = true;
    }

    if (added) {
        // add a handler
        alliances.addNewHandler(new GPHostAllianceHandler(cmd));
    }
}

/** Process alliance record from util.dat.
    \param alliance Alliance container
    \param data Data from util.dat. */
void
GPHostAllianceHandler::processAlliances(GAlliances& alliances, TUtil22Alliance data)
{
    // Alliance offers
    if (GAllianceOffer* pOffer = alliances.getMutableOffer(alliances.find(MAIN_ID))) {
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            pOffer->theirOffer[i] =
                ((data.offered_from[i-1] & MAIN_FLAG) != 0 ? GAllianceOffer::Yes : GAllianceOffer::No);
            pOffer->oldOffer[i] = pOffer->newOffer[i] =
                ((data.offered_to[i-1] & MAIN_FLAG) != 0 ? GAllianceOffer::Yes : GAllianceOffer::No);
        }
    }

    // Level offers
    for (size_t lvl = 0; lvl < NUM_LEVELS; ++lvl) {
        if (GAllianceOffer* pOffer = alliances.getMutableOffer(alliances.find(LEVEL_IDS[lvl]))) {
            for (int i = 1; i <= NUM_PLAYERS; ++i) {
                const uint8_t flag = LEVEL_FLAGS[lvl];
                pOffer->theirOffer[i] =
                    ((data.offered_from[i-1] & flag) != 0
                     ? (data.conditional_from[i-1] & flag) != 0
                     ? GAllianceOffer::Conditional
                     : GAllianceOffer::Yes
                     : GAllianceOffer::No);
                pOffer->oldOffer[i] = pOffer->newOffer[i] =
                    ((data.offered_to[i-1] & flag) != 0
                     ? (data.conditional_to[i-1] & flag) != 0
                     ? GAllianceOffer::Conditional
                     : GAllianceOffer::Yes
                     : GAllianceOffer::No);
            }
        }
    }
}

/** Process enemy record from util.dat.
    \param alliance Alliance container
    \param enemies Flag word from util.dat. */
void
GPHostAllianceHandler::processEnemies(GAlliances& alliances, uint16_t enemies)
{
    // Enemy status
    if (GAllianceOffer* pOffer = alliances.getMutableOffer(alliances.find(ENEMY_ID))) {
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            pOffer->oldOffer[i] = pOffer->newOffer[i] =
                (enemies & (1 << i)) != 0 ? GAllianceOffer::Yes : GAllianceOffer::No;
        }
    }
}
