/**
  *  \file game/specsn.cc
  *  \brief Specification: Short Names
  *
  *  PCC has short names/abbreviations for all units. They are defined
  *  by a specification file "names.cc", supported by PCC since
  *  1.1.12. This module implements short name management for PCC2. It
  *  could also be a home to translated unit names.
  *
  *  Note that this module's interface is intended for use by the
  *  game/spec module only, and not for the general public.
  *
  *  Current implementation: data is loaded into StringTranslator
  *  instances. This has the advantage of being minimally-invasive to
  *  the game/spec module, although it costs quite some runtime
  *  overhead (guesstimated to about 50-100k resident memory).
  *
  *  PCC 1.x implements this feature by reading the file and scanning
  *  for matches, which is nominally O(n*m) during loading, but fast
  *  enough and doesn't need all names in memory.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/specsn.h"
#include "arch/platform.h"
#include "cpluslib/format.h"
#include "util/console.h"
#include "util/filepars.h"
#include "util/misc.h"
#include "util/nls.h"
#include "cpluslib/assert.h"

static const int NUM_TRANSLATORS = 4;

static StringTranslator* translators[NUM_TRANSLATORS];
static const char*const names[NUM_TRANSLATORS] = {
    "HULLS.SHORT",
    "ENGINES.SHORT",
    "BEAMS.SHORT",
    "TORPS.SHORT",
};

namespace {
    class NameFileParser : public FileParser {
        int section;
        void syntaxError(int line_id);
     public:
        NameFileParser();
        void process(int line_id, string_t line);
        void unparsed(int line_id, string_t line);
    };
}

NameFileParser::NameFileParser()
    : FileParser(";#"),
      section(-1)
{ }

void
NameFileParser::syntaxError(int line_id)
{
    console.write(LOG_WARN, format(_("%s:%d: file format error -- line ignored"), getFileName(), line_id));
}

void
NameFileParser::process(int line_id, const string_t line)
{
    string_t::size_type n = line.find_first_not_of(" \t");
    if (n == string_t::npos) {
        // cannot happen
    } else if (line[n] == '[') {
        // section
        string_t::size_type p = line.find(']');
        if (p == string_t::npos) {
            // syntax error
            syntaxError(line_id);
        } else {
            // new section
            const string_t secname(strUCase(string_t(line, n+1, p-n-1)));
            section = -1;
            for (int i = 0; i < NUM_TRANSLATORS; ++i) {
                if (secname == names[i]) {
                    section = i;
                }
            }
        }
    } else if (section >= 0) {
        string_t::size_type p = line.find('=');
        if (p == string_t::npos) {
            // syntax error
            syntaxError(line_id);
        } else {
            // translation
            const string_t name(strUCase(strRTrim(string_t(line, n, p-n))));
            const string_t value(strTrim(string_t(line, p+1)));

            int slot;
            if (stringValue(name, slot)) {
                // numbered entry
                if (translators[section]->getNumberedEntry(slot) == 0) {
                    translators[section]->addNumberedEntry(slot, value.c_str());
                }
            } else {
                // named entry
                const char* name_c = name.c_str();
                if (translators[section]->getTranslation(name_c) == 0) {
                    translators[section]->addTranslation(name_c, value.c_str());
                }
            }
        }
    } else {
        // not in a section
    }
}

void
NameFileParser::unparsed(int /*line_id*/, string_t /*line*/)
{ }


/** Initialize short name module. This loads all the translations. */
void
initShortNames()
{
    // Create and clear all translations
    for (int i = 0; i < NUM_TRANSLATORS; ++i) {
        StringTranslator::getCreateTranslation(names[i], translators[i]);
        translators[i]->clear();
    }

    // Load user files. Language-specific file first
    NameFileParser parser;
    string_t code = Platform::getLanguageCode();
    string_t::size_type n = code.find_first_of("_:-/.\\");
    if (n != code.npos) {
        code.erase(n);
    }
    if (!code.empty()) {
        parser.parseGameFile(format("names_%s.usr", code));
        parser.parseGameFile(format("names_%s.cc", code));
    }

    // System file
    parser.parseGameFile("names.usr");
    parser.parseGameFile("names.cc");
}

/** Get short name.
    \param kind What kind of short name to query
    \param slot Item slot
    \param name Item name */
string_t
getShortName(const TShortName kind, const int slot, const string_t name)
{
    if (kind == snFighterShort) {
        return _("Ftr");
    } else {
        ASSERT(kind < NUM_TRANSLATORS);
        StringTranslator* tx = translators[kind];
        if (tx == 0) {
            return name;
        }
        if (const char* num = tx->getNumberedEntry(slot)) {
            return num;
        }
        if (const char* s = tx->getTranslation(strUCase(name).c_str())) {
            return s;
        }
        return name;
    }
}
