/**
  *  \file tests/movieread.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "tests/movieread.h"
#include "util/except.h"
#include "io/streams.h"

/****************************** MovieReader ******************************/

/** \class MovieReader
    \brief VPM File Reader

    This class provides logic to parse a .vpm movie. The movie file is
    translated into a sequence of virtual function calls. To use it,
    derive a class from MovieReader (or MovieSimpleReader) and
    override the methods.

    You can interrupt reading at block boundaries (new turn, start of
    ship/mine/planet data). */

/** Empty reader. Constructs an empty reader. You must load a stream
    into it before you can use it. */
MovieReader::MovieReader()
    : ip(0)
{ }

MovieReader::~MovieReader()
{ }

/** Load stream. This replaces the movie with the one read from the
    stream, and puts the movie reader into a sane state. */
void
MovieReader::loadFromStream(Stream& s)
{
    data.clear();
    s.readBuffer(data);
    ip = 0;
    newFile();
}

unsigned char
MovieReader::getByte(std::size_t ip)
{
    if (ip >= data.getSize())
        error("data truncated");
    return data[ip];
}

unsigned int
MovieReader::getWord(std::size_t ip)
{
    if (ip+1 >= data.getSize())
        error("data truncated");
    return static_cast<unsigned char>(data[ip])
        + static_cast<unsigned char>(data[ip+1])*256U;
}

/** Reader loop. This reads the movie and calls our virtual functions.
    - when a new turn starts, calls newTurn(); when a new block starts,
      calls newBlock(). If these functions return anything other than
      zero, stops reading and returns that value. If you then call read()
      again, the newTurn()/newBlock() call will be repeated.
    - whenever there's some data, calls one of the changeXXX() functions.
    - this starts with a newTurn() call to go from turn zero to turn one.
    \return zero, or the return value of newTurn()/newBlock().
    \throw FileFormatException when there is a syntactic problem. Override
    error() to change this behaviour. Note that the MovieReader goes into
    an invalid state when it exits through an exception. To get it back
    working, change the position (setPosition()) to a known-good one.
    Known-good positions can be obtained with getPosition() from a
    newTurn()/newBlock() call. */
int
MovieReader::read()
{
    int x;

    /* call 'newTurn' once before going */
    if (ip == 0) {
        if ((x = newTurn()))
            return x;
        ip = START_IP;
    }

    /* normal processing here */
    while (ip < data.getSize()) {
        unsigned char type = getByte(ip);
        if (type == bt_NewTurn) {
            /* new turn */
            if ((x = newTurn()))
                return x;
        } else {
            /* something else */
            if ((x = newBlock(type)))
                return x;
        }
        ++ip;
        switch (type) {
         case bt_NewTurn:
            break;
         case bt_Ship:
            while (unsigned int s = getWord()) {
                unsigned char h      = (s & si_Hull) ? getByte() : 0;
                unsigned char oe     = (s & si_OwnerEngine) ? getByte() : 0;
                unsigned char beam   = (s & si_Beam) ? getByte() : 0;
                unsigned char torp   = (s & si_Torp) ? getByte() : 0;
                unsigned char damage = (s & si_Damage) ? getByte() : 0;
                int x = (s & si_X) ? getWord() : 0;
                int y = (s & si_Y) ? getWord() : 0;
                changeShip((s >> 7) & 511, s, h, oe, beam, torp, damage, x, y);
            }
            break;
         case bt_Planet:
            while (unsigned int p = getWord()) {
                changePlanet(p & 511, (p >> 9) & 15, (p & 0x4000) != 0);
            }
            break;
         case bt_Minefield:
            while (unsigned int mf = getWord()) {
                /* NOTE: VPMC only uses the low 9 bits. We're a bit more
                   flexible to allow for 10k minefields. */
                unsigned int id = mf & 0x7FFFU;
                if (mf & 0x8000U) {
                    changeMinefield(id, getWord());
                } else {
                    int x = getWord(), y = getWord(), radius = getWord(), ot = getByte();
                    changeMinefield(id, x, y, radius, ot);
                }
            }
            break;
         default:
            error("unknown block type");
        }
    }
    return 0;
}

/** Signal an error. This function must not return.
    \default throws a FileFormatException. See read() for more
    information about MovieReader and exceptions. */
void
MovieReader::error(string_t s)
{
    // FIXME: must provide file name!
    throw FileFormatException("<unknown>", s);
}

/** New-file hook. This is called whenever a new movie is loaded into
    this class. */
void
MovieReader::newFile()
{ }

/*************************** MovieSimpleReader ***************************/

/** \class MovieSimpleReader
    \brief Simple VPM Movie reader

    This class implements all MovieReader functions to do nothing. If
    you do not want to override everything, derive your class from
    MovieSimpleReader instead of MovieReader. MovieSimpleReader can
    also be used to check a movie for syntactical correctness (if
    read() does not throw, it's okay).

    In addition, this provides a way to track turn numbers, and to
    stop parsing after turn N. */

/** Constructor.
    \param turn_limit stop parsing after this turn. Omit this
    parameter to read the whole file. */
MovieSimpleReader::MovieSimpleReader(int turn_limit)
    : turn_limit(turn_limit), turn(0)
{ }

int
MovieSimpleReader::newTurn()
{
    if (turn_limit >= 0 && turn >= turn_limit)
        return 1;
    ++turn;
    return 0;
}

int
MovieSimpleReader::newBlock(int /*type*/)
{
    return 0;
}

void
MovieSimpleReader::changeMinefield(int /*id*/, int /*x*/, int /*y*/, int /*radius*/, int /*ot*/)
{ }

void
MovieSimpleReader::changeMinefield(int /*id*/, int /*radius*/)
{ }

void
MovieSimpleReader::changePlanet(int /*id*/, int /*owner*/, bool /*base*/)
{ }

void
MovieSimpleReader::changeShip(int /*id*/, int /*what*/, int /*hull*/, int /*oe*/,
                              int /*beam*/, int /*torp*/, int /*damage*/, int /*x*/, int /*y*/)
{ }
