/**
  *  \file tests/testsim.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <iostream>
#include <string.h>
#include "game/sim.h"
#include "util/console.h"
#include "game/struct.h"
#include "game/gamedirs.h"
#include "game/spec.h"
#include "game/player.h"
#include "game/config.h"
#include "util/except.h"
#include "game/sim-run.h"
#include "game/sim-result.h"
#include "flak/flak.h"
#include "defs.h"
#include "game/racename.h"
#include "util/modinit.h"
#include "arch/platform.h"
#include "game/team.h"

void pullInReferences(FlakBattle& b)
{
#ifdef CONF_FLAK_SUPPORT
    FlakNullVisualizer vis(b);
#endif
}

static void help()
{
    std::cout << "testsim [-opts] file.ccb [directory [count]]\n"
              << "\n"
              << " -p2 -p3 -p4     select PHost 2/3/4 mode\n"
              << " -f              select FLAK mode\n"
              << "                   (default: HOST)\n"
              << " -esb=NN         set engine shield bonus\n"
              << " -[no]scotty     set fed crew bonus\n"
              << " -[no]random     set random left/right\n"
              << " -one -all       just one / all fights\n"
              << " -[no]control    set seed control\n"
              << " -b0 -b360 -bm   select balancing mode to off / 360kt / Master at Arms\n\n";
    std::exit(1);
}

static void tick()
{
//     std::cerr << '.';
//     std::cout << "tick()\n";
}

static void show(const GSimState& st)
{
    for (int i = 0; i < st.getNumShips(); ++i)
        std::cout << "ship " << st.getShip(i).getId()
                  << " '"
                  << st.getShip(i).getName()
                  << "' owned by "
                  << st.getShip(i).getOwner()
                  << ", "
                  << st.getShip(i).getBeamCount()
                  << " beams, "
                  << st.getShip(i).getTorpLauncherCount()
                  << " tubes, "
                  << st.getShip(i).getBayCount()
                  << " bays, "
                  << st.getShip(i).getAmmo()
                  << " ammo"
                  << "\n";
    if (st.hasPlanet()) {
        std::cout << "planet " << st.getPlanet().getId()
                  << " owned by " << st.getPlanet().getOwner()
                  << ", "
                  << st.getPlanet().getDefense()
                  << " dp, "
                  << st.getPlanet().getBaseDefense()
                  << " base defense, beam tech "
                  << st.getPlanet().getBaseBeamTech()
                  << "\n";
    }
}

static const ModuleInitializer* testsim_modules[] = {
    mod_base_console,
    mod_racenames,
    mod_team,
    mod_spec,
    mod_config,
    0
};

int main(int /*argc*/, char** argv)
{
    performInitialisation(OnStartup, testsim_modules);
    argv = Platform::convertArgumentsToUtf8(argv);

    char* file = 0;
    char* dir = 0;
    int count = 1;

    GSimOptions opts;
    
    while (*++argv) {
        if (strcmp(*argv, "-p2") == 0)
            opts.vcr_mode = opts.VcrPHost2;
        else if (strcmp(*argv, "-p3") == 0)
            opts.vcr_mode = opts.VcrPHost4;
        else if (strcmp(*argv, "-p4") == 0)
            opts.vcr_mode = opts.VcrPHost4;
        else if (strcmp(*argv, "-f") == 0)
            opts.vcr_mode = opts.VcrFLAK;
        else if (strncmp(*argv, "-esb=", 5) == 0)
            opts.engine_shield_bonus = atoi(*argv + 5);
        else if (strcmp(*argv, "-scotty") == 0)
            opts.scotty_bonus = true;
        else if (strcmp(*argv, "-noscotty") == 0)
            opts.scotty_bonus = false;
        else if (strcmp(*argv, "-random") == 0)
            opts.random_left_right = true;
        else if (strcmp(*argv, "-norandom") == 0)
            opts.random_left_right = false;
        else if (strcmp(*argv, "-one") == 0)
            opts.only_one_simulation = true;
        else if (strcmp(*argv, "-all") == 0)
            opts.only_one_simulation = false;
        else if (strcmp(*argv, "-control") == 0)
            opts.seed_control = true;
        else if (strcmp(*argv, "-nocontrol") == 0)
            opts.seed_control = false;
        else if (strcmp(*argv, "-b0") == 0)
            opts.balancing_mode = opts.BalanceNone;
        else if (strcmp(*argv, "-b360") == 0)
            opts.balancing_mode = opts.Balance360k;
        else if (strcmp(*argv, "-bm") == 0)
            opts.balancing_mode = opts.BalanceMasterAtArms;
        else if (**argv == '-')
            help();
        else if (!file)
            file = *argv;
        else if (!dir)
            dir = *argv;
        else if (count == 1)
            count = atoi(*argv);
        else
            help();
    }
    if (!file || count <= 0)
        help();

    try {
        initDirectories(dir ? dir : ".", "/usr/games/planets");
        performInitialisation(OnEnterDirectory, testsim_modules);

        GSimResultSummary summary;
        int index = 110;
        for (int ii = 0; ii < count; ++ii) {
            GSimState st;

            Ptr<Stream> s = Stream::openFile(file, Stream::C_READ);
            int file_version;
            char header[6];
            s->readT(header, 6);
            if (memcmp(header, "CCsim\032", 6) == 0) {
                file_version = 0;
            } else if (memcmp(header, "CCbsim", 6) == 0) {
                s->readT(header, 2);
                file_version = (header[0] - '0' + 1);
                if (file_version <= 0 || file_version > 4)
                    throw FileFormatException(*s, "invalid version");
            } else {
                throw FileFormatException(*s, "bad magic");
            }

            /* read it */
            static const unsigned int size_tab[] = { 51, 53, 53, 57, 65 };
            TInt16 count;
            getStructureT(*s, count);
            int num = count.value & 0x7FFF;
            bool planet = (count.value & 0x8000) != 0;

            for (int i = 0; i < num; ++i) {
                char tmp[65];
                memset(tmp, 0, sizeof(tmp));
                s->readT(tmp, size_tab[file_version]);
                TSimShipData sd;
                getStructure(tmp, sd);
                st.getShip(st.addShip()) = GSimShip(sd, file_version);
            }

            /* planet */
            if (planet) {
                char tmp[65];
                memset(tmp, 0, sizeof(tmp));
                s->readT(tmp, size_tab[file_version]);
                TSimPlanetData pd;
                getStructure(tmp, pd);
                st.addPlanet();
                st.getPlanet() = GSimPlanet(pd, file_version);
            }
            std::cout << st.getNumShips() << " ships loaded\n";
            if (st.hasPlanet())
                std::cout << "1 planet loaded\n";

            show(st);

            GSimState orig (st);
            GSimOptions opts;
            opts.vcr_mode            = opts.VcrPHost4;
            opts.seed_control        = true;
            opts.only_one_simulation = true;
            opts.random_left_right   = true;
            opts.balancing_mode      = opts.Balance360k;
            ProgressMonitor monitor;
            monitor.sig_tick.add(tick);
            std::cout << "running it...\n";

            GSimBattleResult res(opts);
            res.init(opts, index);
            index++;
            runSimulation(st, opts, res, monitor);

            std::cout << "result: weight="
                      << res.this_battle_weight
                      << " of "
                      << res.total_battle_weight
                      << ", series "
                      << res.series_length
                      << "\n";
            summary.addResult(orig, st, res);

            if (!res.battles) {
                std::cout << "*** no result!!! ***\n";
            } else {
                std::cout << "result has " << res.battles->getNumBattles() << " battles\n";
                std::cout << "- original:\n";
                show(orig);
                std::cout << "- new state:\n";
                show(st);
                for (int i = 0; i < st.getNumShips(); ++i) {
                    const GSimShip& sh = st.getShip(i);
                    std::cout << "- ship " << i << ": "
                              << sh.getStat().torps_hit << " torps hit, "
                              << orig.getShip(i).getAmmo() - sh.getAmmo() << " lost\n";
                }
            }
        }

        /* output */
        std::cout << "=============== Summary ===============\n";
        for (int i = 0; i < summary.getNumClassResults(); ++i) {
            const GSimResultSummary::ClassResult& cr = summary.getClassResult(i);
            std::cout << "Class " << i << " ("
                      << cr.cumulative_weight << "/" << summary.getCumulativeWeight() << ")\n";
            for (int pl = 1; pl <= 12; ++pl) {
                if (cr.owned_units[pl])
                    std::cout << "  player " << pl << ": " << cr.owned_units[pl] << " units\n";
            }
        }
        for (int i = 0; i < summary.getNumUnitResults(); ++i) {
            const GSimResultSummary::UnitResult& ur = summary.getUnitResult(i);
            std::cout << "Unit " << i << ": "
                      << ur.torps_fired.getTotalScaled() / summary.getCumulativeWeight()
                      << " torps fired, "
                      << ur.fighters_lost.getTotalScaled() / summary.getCumulativeWeight()
                      << " fighters lost\n";
        }
    }
    catch (std::exception& e) {
        std::cerr << e.what() << "\n";
    }
    catch (...) {
        std::cerr << "exception!\n";
    }
}
