/**
  *  \file u/t_game_attachmentconfiguration.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/attachmentunpacker.h"
#include "game/attachmentconfiguration.h"

#include "u/t_game.h"
#include "io/memstream.h"
#include "game/pref.h"


#define UTILDAT_PREFIX_WITH_TIMESTAMP(TIMESTAMP)                        \
    0x0d, 0x00, 0x59, 0x00,                                             \
    TIMESTAMP,  0x0c, 0x00, 0x01, 0x00, 0x04, 0x00, 0x04, 0xda, 0xb0, 0x10, 0xec, 0x94, 0x3d, 0x36, \
    0x04, 0xad, 0xe9, 0x90, 0x38, 0xd4, 0x8d, 0xb7, 0x11, 0x5e, 0xef, 0x6a, 0x0e, 0x79, 0xe8, 0x84, \
    0xc0, 0xbd, 0x6f, 0x03, 0xe7, 0xbe, 0xed, 0xeb, 0x46, 0x4c, 0x41, 0x4b, 0x30, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x6a

/* older timestamp */
#define OLD_TIMESTAMP       0x30, 0x31, 0x2d, 0x30, 0x31, 0x2d, 0x32, 0x30, 0x31, 0x30, 0x31, 0x32, 0x3a, 0x30, 0x30, 0x3a, 0x30, 0x31
#define OLD_UTILDAT_PREFIX  UTILDAT_PREFIX_WITH_TIMESTAMP(OLD_TIMESTAMP)

/* newer timestamp */
#define NEW_TIMESTAMP       0x30, 0x31, 0x2d, 0x30, 0x31, 0x2d, 0x32, 0x30, 0x32, 0x30, 0x31, 0x32, 0x3a, 0x30, 0x30, 0x3a, 0x30, 0x31
#define NEW_UTILDAT_PREFIX  UTILDAT_PREFIX_WITH_TIMESTAMP(NEW_TIMESTAMP)

namespace {
    void clearPreferences()
    {
        GUserPreferences& p = getUserPreferences();
        for (Config::iterator it = p.begin(); it != p.end(); ++it) {
            try {
                it->set("");
            }
            catch (...) { }
            it->setSource(ConfigOption::srcDefault);
        }
    }
}


/** Test attachment configuration, same timestamp again.
    A: create GAttachmentUnpacker. Load util.dat file.
    E: reports UI needed. After saving, reports no UI needed. */
void
GameAttachmentConfigurationTestSuite::testSame()
{
    // Create GAttachmentUnpacker
    GAttachmentUnpacker unpacker;
    clearPreferences();

    // Load util.dat file
    static const uint8_t FILE[] = {
        OLD_UTILDAT_PREFIX,
        34,0, 3+13,0,
        'f','i','l','e','.','t','x','t',0,0,0,0, 0, 'x','y','\n',
    };
    MemoryStream ms(string_t((char*)FILE, sizeof(FILE)));
    TS_ASSERT_THROWS_NOTHING(unpacker.loadUtilData(ms, 1));
    TS_ASSERT_EQUALS(unpacker.getNumAttachments(), 1U);

    // Default configuration says: do user selection
    TS_ASSERT_EQUALS(checkNewAttachments(unpacker), false);

    // Mark processed; configuration now says: skip
    TS_ASSERT_THROWS_NOTHING(markAttachmentsProcessed(unpacker));
    TS_ASSERT_EQUALS(checkNewAttachments(unpacker), true);
}

/** Test attachment configuration, different timestamp.
    A: create GAttachmentUnpacker. Load util.dat file, save it. Create another GAttachmentUnpacker, load another file.
    E: reports UI needed for second GAttachmentUnpacker. */
void
GameAttachmentConfigurationTestSuite::testDifferent()
{
    // Create GAttachmentUnpacker
    clearPreferences();

    {
        // Load first util.dat file
        GAttachmentUnpacker unpacker;
        static const uint8_t FILE[] = {
            OLD_UTILDAT_PREFIX,
            34,0, 3+13,0,
            'f','i','l','e','.','t','x','t',0,0,0,0, 0, 'x','y','\n',
        };
        MemoryStream ms(string_t((char*)FILE, sizeof(FILE)));
        TS_ASSERT_THROWS_NOTHING(unpacker.loadUtilData(ms, 1));
        TS_ASSERT_EQUALS(unpacker.getNumAttachments(), 1U);
        TS_ASSERT_EQUALS(checkNewAttachments(unpacker), false);
        TS_ASSERT_THROWS_NOTHING(markAttachmentsProcessed(unpacker));
    }
    
    {
        // Load second util.dat file
        GAttachmentUnpacker unpacker;
        static const uint8_t FILE[] = {
            NEW_UTILDAT_PREFIX,
            34,0, 3+13,0,
            'f','i','l','e','.','t','x','t',0,0,0,0, 0, 'x','y','\n',
        };
        MemoryStream ms(string_t((char*)FILE, sizeof(FILE)));
        TS_ASSERT_THROWS_NOTHING(unpacker.loadUtilData(ms, 1));
        TS_ASSERT_EQUALS(unpacker.getNumAttachments(), 1U);
        
        // Default configuration says: do user selection
        TS_ASSERT_EQUALS(checkNewAttachments(unpacker), false);
    }
}

/** Test race name handling, "ask" case.
    A: load util.dat containing race names. Configure Unpack.RaceNames=ask.
    E: reports UI needed, attachment selected */
void
GameAttachmentConfigurationTestSuite::testRaceNameDefault()
{
    // Create GAttachmentUnpacker
    GAttachmentUnpacker unpacker;
    clearPreferences();
    getUserPreferences().getConfigOption("Unpack.RaceNames")->set("ask");

    // Load util.dat file
    static const uint8_t FILE[] = {
        OLD_UTILDAT_PREFIX,
        34,0, 3+13,0,
        'r','a','c','e','.','n','m',0,0,0,0,0, 0, 'x','y','\n',
    };
    MemoryStream ms(string_t((char*)FILE, sizeof(FILE)));
    TS_ASSERT_THROWS_NOTHING(unpacker.loadUtilData(ms, 1));
    TS_ASSERT_EQUALS(unpacker.getNumAttachments(), 1U);
    GAttachmentUnpacker::Attachment* att = unpacker.getAttachmentByIndex(0);

    // Verify: do user selection, attachment selected
    TS_ASSERT_EQUALS(checkNewAttachments(unpacker), false);
    TS_ASSERT_EQUALS(unpacker.isAttachmentSelected(att), true);
}

/** Test race name handling, "accept" case.
    A: load util.dat containing race names. Configure Unpack.RaceNames=accept.
    E: reports no UI needed, attachment selected */
void
GameAttachmentConfigurationTestSuite::testRaceNameAccept()
{
    // Create GAttachmentUnpacker
    GAttachmentUnpacker unpacker;
    clearPreferences();
    getUserPreferences().getConfigOption("Unpack.RaceNames")->set("accept");

    // Load util.dat file
    static const uint8_t FILE[] = {
        OLD_UTILDAT_PREFIX,
        34,0, 3+13,0,
        'r','a','c','e','.','n','m',0,0,0,0,0, 0, 'x','y','\n',
    };
    MemoryStream ms(string_t((char*)FILE, sizeof(FILE)));
    TS_ASSERT_THROWS_NOTHING(unpacker.loadUtilData(ms, 1));
    TS_ASSERT_EQUALS(unpacker.getNumAttachments(), 1U);
    GAttachmentUnpacker::Attachment* att = unpacker.getAttachmentByIndex(0);

    // Verify: do user selection, attachment selected
    TS_ASSERT_EQUALS(checkNewAttachments(unpacker), true);
    TS_ASSERT_EQUALS(unpacker.isAttachmentSelected(att), true);
}

/** Test race name handling, "reject" case.
    A: load util.dat containing race names. Configure Unpack.RaceNames=reject.
    E: reports no UI needed, attachment unselected */
void
GameAttachmentConfigurationTestSuite::testRaceNameReject()
{
    // Create GAttachmentUnpacker
    GAttachmentUnpacker unpacker;
    clearPreferences();
    getUserPreferences().getConfigOption("Unpack.RaceNames")->set("reject");

    // Load util.dat file
    static const uint8_t FILE[] = {
        OLD_UTILDAT_PREFIX,
        34,0, 3+13,0,
        'r','a','c','e','.','n','m',0,0,0,0,0, 0, 'x','y','\n',
    };
    MemoryStream ms(string_t((char*)FILE, sizeof(FILE)));
    TS_ASSERT_THROWS_NOTHING(unpacker.loadUtilData(ms, 1));
    TS_ASSERT_EQUALS(unpacker.getNumAttachments(), 1U);
    GAttachmentUnpacker::Attachment* att = unpacker.getAttachmentByIndex(0);

    // Verify: do user selection, attachment selected
    TS_ASSERT_EQUALS(checkNewAttachments(unpacker), true);
    TS_ASSERT_EQUALS(unpacker.isAttachmentSelected(att), false);
}

/** Test empty GAttachmentUnpacker.
    A: create empty GAttachmentUnpacker.
    E: returns no UI needed. */
void
GameAttachmentConfigurationTestSuite::testEmpty()
{
    // Create GAttachmentUnpacker
    GAttachmentUnpacker unpacker;
    TS_ASSERT_EQUALS(checkNewAttachments(unpacker), true);
}

