/**
  *  \file ui/richtext.cc
  *  \brief Text with Attributes
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/richtext.h"
#include "util/misc.h"

/************************* RichTextColorAttribute ************************/

/** Create color attribute. */
RichTextColorAttribute::RichTextColorAttribute(UIColor::Color color)
    : color(color)
{ }

RichTextColorAttribute::~RichTextColorAttribute()
{ }

RichTextColorAttribute*
RichTextColorAttribute::clone() const
{
    return new RichTextColorAttribute(color);
}

/************************* RichTextStyleAttribute ************************/

RichTextStyleAttribute::RichTextStyleAttribute(Style style)
    : style(style)
{ }

RichTextStyleAttribute::~RichTextStyleAttribute()
{ }

RichTextStyleAttribute*
RichTextStyleAttribute::clone() const
{
    return new RichTextStyleAttribute(style);
}

/************************* RichTextLinkAttribute *************************/

RichTextLinkAttribute::RichTextLinkAttribute(const string_t target)
    : target(target)
{ }

RichTextLinkAttribute::~RichTextLinkAttribute()
{ }

RichTextLinkAttribute*
RichTextLinkAttribute::clone() const
{
    return new RichTextLinkAttribute(target);
}


/******************************** RichText *******************************/

/*! \class RichText
    \brief Text with Attributes

    This class is similar to a std::string, but in addition allows to
    associate arbitrary attributes with runs of characters. Attributes
    can be nested, but there must be a subset ordering on the ranges:
    if two ranges overlap, one must be a subset of the other. With
    this behaviour, RichText is similar to XML tags.

    Attributes are descendants of RichTextAttribute. For convenience,
    a RichTextColorAttribute is provided and explicitly supported by
    our constructors.

    In addition to method calls, you can use RichTextVisitor to
    inspect a RichText object. */

/** Construct blank object. */
RichText::RichText()
{ }

/** Construct from C string. Creates unattributed text. */
RichText::RichText(const char* text)
    : raw_text(text)
{ }

/** Construct from C++ string. Creates unattributed text. */
RichText::RichText(string_t text)
    : raw_text(text)
{ }

/** Construct colored text from C string.
    \param color Color to assign to whole string
    \param text  Text */
RichText::RichText(UIColor::Color color, const char* text)
    : raw_text(text)
{
    withAttribute(new RichTextColorAttribute(color));
}

/** Construct colored text from C++ string.
    \param color Color to assign to whole string
    \param text  Text */
RichText::RichText(UIColor::Color color, string_t text)
    : raw_text(text)
{
    withAttribute(new RichTextColorAttribute(color));
}

/** Copy-Constructor. */
RichText::RichText(const RichText& other)
    : raw_text(other.raw_text)
{
    attributes.reserve(other.attributes.size());
    for (attindex_t i = 0; i < other.attributes.size(); ++i) {
        RichTextAttribute* att = other.attributes[i]->clone();
        att->start = other.attributes[i]->start;
        att->end = other.attributes[i]->end;
        attributes.push_back(att);
    }
}

/** Construct sub-string.
    \param other  Rich-text object to copy from
    \param start  Start at this position
    \param length Copy this many characters */
RichText::RichText(const RichText& other, size_type start, size_type length)
    : raw_text(other.raw_text, start, length)
{
    const size_type end = length > string_t::npos - start ? string_t::npos : start + length;
    for (attindex_t i = 0; i < other.attributes.size(); ++i) {
        const RichTextAttribute* attr = other.attributes[i];
        /* starting position of new attribute */
        size_type nstart = attr->start > start ? attr->start : start;
        /* ending position of new attribute */
        size_type nend = attr->end < end ? attr->end : end;
        /* attribute effective in this section? */
        if (nstart < nend) {
            RichTextAttribute* nattr = attr->clone();
            nattr->start = nstart - start;
            nattr->end   = nend - start;
            attributes.push_back(nattr);
        }
    }
}

/** Destructor. */
RichText::~RichText()
{ }

/** Apply attribute to whole text.
    \param attr attribute to apply. This object takes ownership of the
                object. On error, deletes the object.
    \return *this */
RichText&
RichText::withAttribute(RichTextAttribute* attr)
{
    if (raw_text.size()) {
        attr->start = 0;
        attr->end   = raw_text.size();
        attributes.insert(attributes.begin(), attr);
    } else {
        delete attr;
    }
    return *this;
}

/** Apply attribute to whole text. Shortcut to apply a RichTextStyleAttribute. */
RichText&
RichText::withAttribute(RichTextStyleAttribute::Style style)
{
    return withAttribute(new RichTextStyleAttribute(style));
}

/** Get raw text without attributes. */
string_t
RichText::getText() const
{
    return raw_text;
}

/** Get substring of a rich-text object.
    \param start  first position
    \param length number of characters to extract */
RichText
RichText::substr(size_type start, size_type length) const
{
    return RichText(*this, start, length);
}

/** Erase part of a rich-text object. */
void
RichText::erase(size_type start, size_type length)
{
    if (start < raw_text.size()) {
        if (length >= raw_text.size() - start) {
            /* delete till eos */
            *this = RichText(*this, 0, start);
        } else {
            /* delete piece from middle */
            RichText tmp(*this, 0, start);
            tmp.append(RichText(*this, start + length, string_t::npos));
            *this = tmp;
        }
    }
}

/** Append other rich text object.
    \param other [in] text to append
    \return *this */
RichText&
RichText::append(const RichText& other)
{
    attindex_t offset = raw_text.size();
    raw_text.append(other.raw_text);
    for (attindex_t i = 0; i < other.attributes.size(); ++i) {
        RichTextAttribute* att = other.attributes[i]->clone();
        att->start = other.attributes[i]->start + offset;
        att->end   = other.attributes[i]->end   + offset;
        attributes.push_back(att);
    }
    return *this;
}

/** Append attribute-less text.
    \param text [in] text to append
    \return *this */
RichText&
RichText::append(const char* text)
{
    raw_text.append(text);
    return *this;
}

/** Append attribute-less text.
    \param text [in] text to append
    \return *this */
RichText&
RichText::append(string_t text)
{
    raw_text.append(text);
    return *this;
}

/** Append colored text.
    \param color [in] color to use
    \param text [in] text to append
    \return *this */
RichText&
RichText::append(UIColor::Color color, const char* text)
{
    size_type start = raw_text.size();
    raw_text.append(text);

    RichTextAttribute* att = new RichTextColorAttribute(color);
    att->start = start;
    att->end = raw_text.size();
    attributes.push_back(att);
    return *this;
}

/** Append colored text.
    \param color [in] color to use
    \param text [in] text to append
    \return *this */
RichText&
RichText::append(UIColor::Color color, string_t text)
{
    size_type start = raw_text.size();
    raw_text.append(text);

    RichTextAttribute* att = new RichTextColorAttribute(color);
    att->start = start;
    att->end = raw_text.size();
    attributes.push_back(att);
    return *this;
}

/** Assignment operator. */
RichText&
RichText::operator=(const RichText& other)
{
    if (&other != this) {
        RichText tmp(other);
        tmp.swap(*this);
    }
    return *this;
}

/** Swap two rich-text objects. */
void
RichText::swap(RichText& other)
{
    raw_text.swap(other.raw_text);
    attributes.swap(other.attributes);
}

/** Clear this rich-text object. */
void
RichText::clear()
{
    raw_text.clear();
    attributes.clear();
}

/** Visit this text item. */
RichTextVisitor&
RichText::visit(RichTextVisitor& visitor) const
{
    size_type index = 0;
    std::vector<RichTextAttribute*> att_stack;
    attindex_t att_index = 0;

    while (index < raw_text.size()) {
        size_type next_event = raw_text.size();
        if (att_stack.size() && att_stack.back()->end < next_event)
            next_event = att_stack.back()->end;
        if (att_index < attributes.size() && attributes[att_index]->start < next_event)
            next_event = attributes[att_index]->start;

        if (next_event > index) {
            /* we have some text */
            if (!visitor.handleText(raw_text.substr(index, next_event - index)))
                goto out;
        } else {
            /* leave some attributes */
            while (att_stack.size() && att_stack.back()->end <= index) {
                if (!visitor.endAttribute(att_stack.back()))
                    goto out;
                att_stack.pop_back();
            }

            /* enter some attributes */
            while (att_index < attributes.size() && attributes[att_index]->start <= index) {
                if (!visitor.startAttribute(attributes[att_index]))
                    goto out;
                att_stack.push_back(attributes[att_index]);
                ++att_index;
            }
        }

        index = next_event;
    }

    /* just to be sure... */
    while (att_stack.size()) {
        if (!visitor.endAttribute(att_stack.back()))
            goto out;
        att_stack.pop_back();
    }

 out:
    return visitor;
}


RichText
operator+(const RichText& lhs, const RichText& rhs)
{
    RichText tmp(lhs);
    tmp += rhs;
    return tmp;
}

RichText
operator+(const char* lhs, const RichText& rhs)
{
    RichText tmp(lhs);
    tmp += rhs;
    return tmp;
}

RichText
operator+(const RichText& lhs, const char* rhs)
{
    RichText tmp(lhs);
    tmp += rhs;
    return tmp;
}
