/**
  *  \file ui/richtext.h
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_RICHTEXT_H
#define PCC_V2_UI_RICHTEXT_H

#include <vector>
#include "ui/skin.h"
#include "cpluslib/string.h"
#include "cpluslib/ptrvector.h"

/* Attributes:
   Color
   Hyperlink
   Font */

class RichText;
class RichTextVisitor;

class RichTextAttribute {
    friend class RichText;
    string_t::size_type start, end;
 public:
    RichTextAttribute()
        : start(0), end(0)
        { }
    virtual ~RichTextAttribute()
        { }
    virtual RichTextAttribute* clone() const = 0;
};

class RichTextColorAttribute : public RichTextAttribute {
    UIColor::Color color;
 public:
    RichTextColorAttribute(UIColor::Color color);
    ~RichTextColorAttribute();
    RichTextColorAttribute* clone() const;

    UIColor::Color getColor() const
        { return color; }
};

class RichTextStyleAttribute : public RichTextAttribute {
 public:
    enum Style {
        Bold,
        Italic,
        Underline,
        Big,
        Small,
        Fixed,
        Key
    };
    RichTextStyleAttribute(Style style);
    ~RichTextStyleAttribute();
    RichTextStyleAttribute* clone() const;
    Style getStyle() const
        { return style; }
 private:
    Style style;
};

class RichTextLinkAttribute : public RichTextAttribute {
    string_t target;
 public:
    RichTextLinkAttribute(const string_t target);
    ~RichTextLinkAttribute();
    RichTextLinkAttribute* clone() const;

    const string_t& getTarget() const
        { return target; }
};

class RichText {
    /** The raw text. */
    string_t raw_text;
    /** All attributes, sorted by start then end. Note that attributes
        cannot overlap. */
    ptr_vector<RichTextAttribute> attributes;
    typedef ptr_vector<RichTextAttribute>::size_type attindex_t;
 public:
    typedef string_t::size_type size_type;
    
    RichText();
    RichText(const char* text);
    RichText(string_t text);
    RichText(UIColor::Color color, const char* text);
    RichText(UIColor::Color color, string_t text);
    RichText(const RichText& other);
    RichText(const RichText& other, size_type start, size_type length = string_t::npos);
    ~RichText();

    RichText& withAttribute(RichTextAttribute* attr);
    RichText& withAttribute(RichTextStyleAttribute::Style style);

    string_t getText() const;

    /** Get number of attribute records. */
    int getNumAttributes() const
        { return attributes.size(); }

    RichText substr(size_type start, size_type length = string_t::npos) const;
    void erase(size_type start, size_type length = string_t::npos);

    string_t::size_type find(char what, string_t::size_type start_at = 0) const
        { return raw_text.find(what, start_at); }

    RichText& append(const RichText& other);
    RichText& append(const char* text);
    RichText& append(string_t text);
    RichText& append(UIColor::Color color, const char* text);
    RichText& append(UIColor::Color color, string_t text);

    RichText& operator=(const RichText& other);
    RichText& operator+=(const RichText& other)
        { return append(other); }
    RichText& operator+=(const char* other)
        { return append(other); }
    RichText& operator+=(string_t text)
        { return append(text); }

    size_type size() const
        { return raw_text.size(); }
    size_type length() const
        { return raw_text.size(); }
    bool empty() const
        { return raw_text.empty(); }
    char operator[](size_type i) const
        { return raw_text[i]; }

    void swap(RichText& other);
    void clear();

    RichTextVisitor& visit(RichTextVisitor& visitor) const;
};

class RichTextVisitor {
 public:
    RichTextVisitor()
        { }
    virtual ~RichTextVisitor()
        { }

    /** Handle run of text.
        \param text Text
        \return true to continue iteration, false to exit. */
    virtual bool handleText(string_t text) = 0;

    /** Handle beginning of an attribute.
        \param att the attribute
        \return true to continue iteration, false to exit. */
    virtual bool startAttribute(const RichTextAttribute* att) = 0;

    /** Handle end of an attribute. Note that this endAttribute()
        always refers to the last not-yet-ended startAttribute().
        \param att the attribute
        \return true to continue iteration, false to exit. */
    virtual bool endAttribute(const RichTextAttribute* att) = 0;

    RichTextVisitor& visit(const RichText& text)
        { return text.visit(*this); }
};

RichText operator+(const RichText& lhs, const RichText& rhs);
RichText operator+(const char* lhs, const RichText& rhs);
RichText operator+(const RichText& lhs, const char* rhs);

#endif
